"""A core package, containing all the basic class and functions that serves as
    the foundation of Dualing common modules.
"""

from dualing.core.dataset import Dataset
from dualing.core.loss import (BinaryCrossEntropy, ContrastiveLoss,
                               TripletHardLoss, TripletSemiHardLoss)
from dualing.core.model import Base, Siamese



"""Triplet Loss Siamese Network.
"""

import tensorflow as tf

import dualing.utils.exception as e
import dualing.utils.logging as l
from dualing.core import Siamese, TripletHardLoss, TripletSemiHardLoss

logger = l.get_logger(__name__)


class TripletSiamese(Siamese):
    """A TripletSiamese class is responsible for implementing the
    triplet-loss version of Siamese Neural Networks.

    References:
        X. Dong and J. Shen. Triplet loss in siamese network for object tracking.
        Proceedings of the European Conference on Computer Vision (2018).

    """

    def __init__(self, base, loss='hard', margin=1.0, soft=False, distance_metric='L2', name=''):
        """Initialization method.

        Args:
            base (Base): Twin architecture.
            loss (str): Whether network should use hard or semi-hard negative mining.
            margin (float): Radius around the embedding space.
            soft (bool): Whether network should use soft margin or not.
            distance_metric (str): Distance metric.
            name (str): Naming identifier.

        """

        logger.info('Overriding class: Siamese -> TripletSiamese.')

        # Overrides its parent class with any custom arguments if needed
        super(TripletSiamese, self).__init__(base, name=name)

        # Type of loss
        self.loss_type = loss

        # Radius around embedding space
        self.margin = margin

        # Soft margin
        self.soft = soft

        # Distance metric
        self.distance = distance_metric

        logger.info('Class overrided.')

    @property
    def loss_type(self):
        """str: Type of loss (hard or semi-hard).

        """

        return self._loss_type

    @loss_type.setter
    def loss_type(self, loss_type):
        if loss_type not in ['hard', 'semi-hard']:
            raise e.ValueError('`loss_type` should be `hard` or `semi-hard`')

        self._loss_type = loss_type

    @property
    def soft(self):
        """bool: Whether soft margin should be used or not.

        """

        return self._soft

    @soft.setter
    def soft(self, soft):
        if not isinstance(soft, bool):
            raise e.TypeError('`soft` should be a boolean')

        self._soft = soft

    @property
    def margin(self):
        """float: Radius around the embedding space.

        """

        return self._margin

    @margin.setter
    def margin(self, margin):
        if not isinstance(margin, float):
            raise e.TypeError('`margin` should be a float')
        if margin <= 0:
            raise e.ValueError('`margin` should be greater than 0')

        self._margin = margin

    @property
    def distance(self):
        """str: Distance metric.

        """

        return self._distance

    @distance.setter
    def distance(self, distance):
        if distance not in ['L1', 'L2', 'angular']:
            raise e.ValueError('`distance` should be `L1`, `L2` or `angular`')

        self._distance = distance

    def compile(self, optimizer):
        """Method that builds the network by attaching optimizer, loss and metrics.

        Args:
            optimizer (tf.keras.optimizers): Optimization algorithm.

        """

        # Creates an optimizer object
        self.optimizer = optimizer

        # Check it is supposed to use hard negative mining
        if self.loss_type == 'hard':
            # Creates the TripletHardLoss
            self.loss = TripletHardLoss()

        # If it is supposed to use semi-hard negative mining
        elif self.loss_type == 'semi-hard':
            # Creates the TripletSemiHardLoss
            self.loss = TripletSemiHardLoss()

        # Defines the loss metric
        self.loss_metric = tf.metrics.Mean(name='loss')

    @tf.function
    def step(self, x, y):
        """Method that performs a single batch optimization step.

        Args:
            x (tf.Tensor): Tensor containing samples.
            y (tf.Tensor): Tensor containing labels.

        """

        # Uses tensorflow's gradient
        with tf.GradientTape() as tape:
            # Passes the batch inputs through the network
            y_pred = self.B(x)

            # Performs the L2 normalization prior to the loss function
            y_pred = tf.math.l2_normalize(y_pred, axis=-1)

            # Calculates the loss
            loss = self.loss(y, y_pred, self.margin, self.soft, self.distance)

        # Calculates the gradients for each training variable based on the loss function
        gradients = tape.gradient(loss, self.B.trainable_variables)

        # Applies the gradients using an optimizer
        self.optimizer.apply_gradients(
            zip(gradients, self.B.trainable_variables))

        # Updates the metrics' states
        self.loss_metric.update_state(loss)

    def fit(self, batches, epochs=100):
        """Method that trains the model over training batches.

        Args:
            batches (BatchDataset): Batches of tuples holding training samples and labels.
            epochs (int): Maximum number of epochs.

        """

        logger.info('Fitting model ...')

        # Gathers the amount of batches
        n_batches = tf.data.experimental.cardinality(batches).numpy()

        # Iterates through all epochs
        for epoch in range(epochs):
            logger.info('Epoch %d/%d', epoch+1, epochs)

            # Resets metrics' states
            self.loss_metric.reset_states()

            # Defines a customized progress bar
            b = tf.keras.utils.Progbar(n_batches, stateful_metrics=['loss'])

            # Iterates through all batches
            for (x_batch, y_batch) in batches:
                # Performs the optimization step
                self.step(x_batch, y_batch)

                # Adds corresponding values to the progress bar
                b.add(1, values=[('loss', self.loss_metric.result())])

            logger.file(f'Loss: {self.loss_metric.result()}')

    def evaluate(self, batches):
        """Method that evaluates the model over validation or testing batches.

        Args:
            batches (BatchDataset): Batches of tuples holding validation / test samples and labels.

        """

        logger.info('Evaluating model ...')

        # Gathers the amount of batches
        n_batches = tf.data.experimental.cardinality(batches).numpy()

        # Resets metrics' states
        self.loss_metric.reset_states()

        # Defines a customized progress bar
        b = tf.keras.utils.Progbar(n_batches, stateful_metrics=['val_loss'])

        # Iterates through all batches
        for (x_batch, y_batch) in batches:
            # Performs the prediction
            y_pred = self.predict(x_batch)

            # Calculates the loss
            loss = self.loss(y_batch, y_pred, self.margin)

            # Updates the metrics' states
            self.loss_metric.update_state(loss)

            # Adds corresponding values to the progress bar
            b.add(1, values=[('val_loss', self.loss_metric.result())])

        logger.file(f'Val Loss: {self.loss_metric.result()}')

    def predict(self, x):
        """Method that performs a forward pass over samples and returns the network's output.

        Args:
            x (tf.Tensor): Tensor containing input samples.

        Returns:
            The embedded and L2 normalized data from `x`.

        """

        # Passes the sample through the network
        y_pred = self.B(x)

        # Performs the L2 normalization prior to the loss function
        y_pred = tf.math.l2_normalize(y_pred, axis=-1)

        return y_pred



"""Dataset.
"""

import tensorflow as tf

import dualing.utils.exception as e
import dualing.utils.logging as l

logger = l.get_logger(__name__)


class Dataset:
    """A Dataset class is responsible for receiving raw data, pre-processing it and
    persisting batches that will be feed as inputs to the networks.

    """

    def __init__(self, batch_size=1, input_shape=None, normalize=(0, 1), shuffle=True, seed=0):
        """Initialization method.

        Args:
            batch_size (int): Batch size.
            input_shape (tuple): Shape of the reshaped array.
            normalize (tuple): Normalization bounds.
            shuffle (bool): Whether data should be shuffled or not.
            seed (int): Provides deterministic traits when using `random` module.

        """

        # Batch size
        self.batch_size = batch_size

        # Shape of the input tensors
        self.input_shape = input_shape

        # Normalization bounds
        self.normalize = normalize

        # Whether data should be shuffled or not
        self.shuffle = shuffle

        # Defines the tensorflow random seed
        tf.random.set_seed(seed)

        # Debugs important information
        logger.debug('Size: %s | Batch size: %s | Normalization: %s | Shuffle: %s.',
                     input_shape, batch_size, normalize, shuffle)

    @property
    def batch_size(self):
        """int: Batch size.

        """

        return self._batch_size

    @batch_size.setter
    def batch_size(self, batch_size):
        if not isinstance(batch_size, int):
            raise e.TypeError('`batch_size` should be a integer')
        if batch_size <= 0:
            raise e.ValueError('`batch_size` should be greater than 0')

        self._batch_size = batch_size

    @property
    def input_shape(self):
        """tuple: Shape of the input tensors.

        """

        return self._input_shape

    @input_shape.setter
    def input_shape(self, input_shape):
        if not (isinstance(input_shape, tuple) or input_shape is None):
            raise e.TypeError('`input_shape` should be a tuple or None')

        self._input_shape = input_shape

    @property
    def normalize(self):
        """tuple: Normalization bounds.

        """

        return self._normalize

    @normalize.setter
    def normalize(self, normalize):
        if not (isinstance(normalize, tuple) or normalize is None):
            raise e.TypeError('`normalize` should be a tuple or None')

        self._normalize = normalize

    @property
    def shuffle(self):
        """bool: Whether data should be shuffled or not.

        """

        return self._shuffle

    @shuffle.setter
    def shuffle(self, shuffle):
        if not isinstance(shuffle, bool):
            raise e.TypeError('`shuffle` should be a boolean')

        self._shuffle = shuffle

    def preprocess(self, data):
        """Pre-process the data by reshaping and normalizing, if necessary.

        Args:
            data (np.array): Array of data.

        Returns:
            Pre-processed data.

        """

        # If a shape is supplied
        if self.input_shape:
            # Reshapes the array and make sure that it is `float`
            data = data.reshape(self.input_shape).astype('float32')

        # If no shape is supplied
        else:
            # Just make sure that the array is `float`
            data = data.astype('float32')

        # If data should be normalized
        if self.normalize:
            # Gathers the lower and upper bounds of normalization
            low, high = self.normalize[0], self.normalize[1]

            # Gathers the minimum and maximum values of the data
            _min, _max = tf.math.reduce_min(data), tf.math.reduce_max(data)

            # Normalizes the data between `low` and `high`
            data = (high - low) * ((data - _min) / (_max - _min)) + low

        return data

    def _build(self):
        """Method that builds the class.

        Note that you need to implement this method directly on its child. Essentially,
        each Dataset has its building procedure.

        Raises:
            NotImplementedError.

        """

        raise NotImplementedError



"""Logging-based methods and helpers.
"""

import logging
import sys
from logging.handlers import TimedRotatingFileHandler

FORMATTER = logging.Formatter(
    '%(asctime)s - %(name)s — %(levelname)s — %(message)s')
LOG_FILE = 'dualing.log'
LOG_LEVEL = logging.DEBUG


class Logger(logging.Logger):
    """A customized Logger file that enables the possibility of only logging to file.

    """

    def file(self, msg, *args, **kwargs):
        """Logs the message only to the logging file.

        Args:
            msg (str): Message to be logged.

        """

        # Sets the console handler as critical level to disable console logging
        self.handlers[0].setLevel(logging.CRITICAL)

        # Logs the information
        self.info(msg, *args, **kwargs)

        # Re-enables the console handler logging
        self.handlers[0].setLevel(LOG_LEVEL)


def get_console_handler():
    """Gets a console handler to handle logging into console.

    Returns:
        A handler to output information into console.

    """

    # Creates a stream handler for logger
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(FORMATTER)

    return console_handler


def get_timed_file_handler():
    """Gets a timed file handler to handle logging into files.

    Returns:
        A handler to output information into timed files.

    """

    # Creates a timed rotating file handler for logger
    file_handler = TimedRotatingFileHandler(LOG_FILE, when='midnight')
    file_handler.setFormatter(FORMATTER)

    return file_handler


def get_logger(logger_name):
    """Gets a log and make it avaliable for further use.

    Args:
        logger_name (str): The name of the logger.

    Returns:
        A handler to output information into console's.

    """

    # Defining a customized logger in order to have the possibility
    # of only logging to file when desired
    logging.setLoggerClass(Logger)

    # Creates a logger object
    logger = logging.getLogger(logger_name)

    # Sets an log level
    logger.setLevel(LOG_LEVEL)

    # Adds the desired handlers
    logger.addHandler(get_console_handler())
    logger.addHandler(get_timed_file_handler())

    # True or False for propagating logs
    logger.propagate = False

    return logger



"""Batch-based dataset.
"""

import tensorflow as tf

import dualing.utils.constants as c
import dualing.utils.exception as e
import dualing.utils.logging as l
from dualing.core import Dataset

logger = l.get_logger(__name__)


class BatchDataset(Dataset):
    """A BatchDataset class is responsible for implementing a standard dataset
    that uses input data and labels to provide batches.

    """

    def __init__(self, data, labels, batch_size=1, input_shape=None, normalize=(0, 1),
                 shuffle=True, seed=0):
        """Initialization method.

        Args:
            data (np.array): Array of samples.
            labels (np.array): Array of labels.
            batch_size (int): Batch size.
            input_shape (tuple): Shape of the reshaped array.
            normalize (tuple): Normalization bounds.
            shuffle (bool): Whether data should be shuffled or not.
            seed (int): Provides deterministic traits when using `random` module.

        """

        logger.info('Overriding class: Dataset -> BatchDataset.')

        # Overrides its parent class with any custom arguments if needed
        super(BatchDataset, self).__init__(
            batch_size, input_shape, normalize, shuffle, seed)

        # Pre-processes the data
        data = self.preprocess(data)

        # Builds up the class
        self._build(data, labels)

        logger.info('Class overrided.')

    @property
    def batches(self):
        """tf.data.Dataset: Batches of data (samples, labels).

        """

        return self._batches

    @batches.setter
    def batches(self, batches):
        if not isinstance(batches, tf.data.Dataset):
            raise e.TypeError('`batches` should be a tf.data.Dataset')

        self._batches = batches

    def _build(self, data, labels):
        """Builds the class.

        Args:
            data (np.array): Array of samples.
            labels (np.array): Array of labels.

        """

        # Checks if data should be shuffled
        if self.shuffle:
            # Creates dataset from shuffled and batched data
            self.batches = tf.data.Dataset.from_tensor_slices(
                (data, labels)).shuffle(c.BUFFER_SIZE).batch(self.batch_size)

        # If data should not be shuffled
        else:
            # Creates dataset from batched data
            self.batches = tf.data.Dataset.from_tensor_slices(
                (data, labels)).batch(self.batch_size)



"""A package for already-implemented machine learning Siamese architectures.
"""

from dualing.models.contrastive import ContrastiveSiamese
from dualing.models.cross_entropy import CrossEntropySiamese
from dualing.models.triplet import TripletSiamese



"""Convolutional Neural Network.
"""

from tensorflow.keras.layers import Conv2D, Dense, Flatten, MaxPool2D

import dualing.utils.logging as l
from dualing.core import Base

logger = l.get_logger(__name__)


class CNN(Base):
    """A CNN class stands for a standard Convolutional Neural Network implementation.

    """

    def __init__(self, n_blocks=3, init_kernel=5, n_output=128, activation='sigmoid'):
        """Initialization method.

        """

        logger.info('Overriding class: Base -> CNN.')

        # Overrides its parent class with any custom arguments if needed
        super(CNN, self).__init__(name='cnn')

        # Asserting that it will be possible to create the convolutional layers
        assert init_kernel - 2 * (n_blocks - 1) >= 1

        # Convolutional layers
        self.conv = [Conv2D(32 * (2 ** i), init_kernel - 2 * i, activation='relu', padding='same')
                     for i in range(n_blocks)]

        # Pooling layers
        self.pool = [MaxPool2D() for _ in range(n_blocks)]

        # Flatenning layer
        self.flatten = Flatten()

        # Final fully-connected layer
        self.fc = Dense(n_output, activation=activation)

        logger.info('Class overrided.')
        logger.debug('Blocks: %d| Initial Kernel: %d | Output (%s): %d', n_blocks, init_kernel, activation, n_output)

    def call(self, x):
        """Method that holds vital information whenever this class is called.

        Args:
            x (tf.Tensor): Tensor containing the input sample.

        Returns:
            The layer's outputs.

        """

        # Iterate through convolutional and poooling layers
        for (conv, pool) in zip(self.conv, self.pool):
            # Pass through convolutional layer
            x = conv(x)

            # Pass through pooling layer
            x = pool(x)

        # Flattens the outputs
        x = self.flatten(x)

        # Pass through the fully-connected layer
        x = self.fc(x)

        return x



"""Helpers and methods to project embedded data.
"""

import matplotlib.pyplot as plt
import numpy as np
import tensorflow as tf


def _tensor_to_numpy(tensor):
    """Converts a tensor to a numpy array.

    Args:
        tensor (tf.Tensor): Tensor to be converted.

    Returns:
        A numpy array with the same values as the tensor.

    """

    # Checks if the inputted tensor is really a tensor
    if tf.is_tensor(tensor):
        # If yes, returns its numpy version
        return tensor.numpy()

    # If no, just returns it
    return tensor


def plot_embeddings(embeddings, labels, dims=(0, 1)):
    """Plots embedded data along their true labels.

    Args:
        embeddings (tf.Tensor, np.array): Tensor or array holding the embedded data.
        labels (tf.Tensor, np.array): Tensor or array holding the true labels.
        dims (tuple): Dimensions to be plotted.

    """

    # Makes sure that embeddings will be a numpy array
    embeddings = _tensor_to_numpy(embeddings)

    # Makes sure that labels will be a numpy array
    labels = _tensor_to_numpy(labels)

    # Creates figure and axis subplots
    _, axis = plt.subplots(figsize=(13, 7))

    # Creates the axis labels strings
    x_label, y_label = f'$x_{dims[0]}$', f'$x_{dims[1]}$'

    # Defines some properties, such as axis labels
    axis.set(xlabel=r'{}'.format(x_label), ylabel=r'{}'.format(y_label))

    # Iterates through every possible labels
    for i in range(np.max(labels) + 1):
        # Gathers the indexes
        indexes = np.where(labels == i)[0]

        # Scatter plots the desired dimensions (2-D)
        plt.scatter(embeddings[indexes, dims[0]],
                    embeddings[indexes, dims[1]], alpha=0.75, label=i)

    # Adds a legend to the plot
    plt.legend()

    # Shows the plot
    plt.show()



"""Base architecture and Siamese Network.
"""

import tensorflow as tf

import dualing.utils.exception as e


class Base(tf.keras.Model):
    """A Base class is responsible for easily-implementing the
    base twin architecture of a Siamese Network.

    """

    def __init__(self, name=''):
        """Initialization method.

        Args:
            name (str): Naming identifier.

        """

        # Overrides its parent class with any custom arguments if needed
        super(Base, self).__init__(name=name)

    def call(self, x):
        """Method that holds vital information whenever this class is called.

        Note that you need to implement this method directly on its child. Essentially,
        each neural network has its own forward pass implementation.

        Args:
            x (tf.Tensor): Tensor containing the input sample.

        Raises:
            NotImplementedError.

        """

        raise NotImplementedError


class Siamese(tf.keras.Model):
    """An Siamese class is responsible for implementing the base of Siamese Neural Networks.

    """

    def __init__(self, base, name=''):
        """Initialization method.

        Args:
            base (Base): Twin architecture.
            name (str): Naming identifier.

        """

        # Overrides its parent class with any custom arguments if needed
        super(Siamese, self).__init__(name=name)

        # Defines the Siamese's base twin architecture
        self.B = base

    @property
    def B(self):
        """Base: Twin architecture.

        """

        return self._B

    @B.setter
    def B(self, B):
        if not isinstance(B, Base):
            raise e.TypeError('`B` should be a child from Base class')

        self._B = B

    def compile(self, optimizer):
        """Method that builds the network by attaching optimizer, loss and metrics.

        Note that you need to implement this method directly on its child. Essentially,
        each type of Siamese has its own set of loss and metrics.

        Args:
            optimizer (tf.keras.optimizers): Optimization algorithm.

        Raises:
            NotImplementedError.

        """

        raise NotImplementedError

    def step(self, x, y):
        """Method that performs a single batch optimization step.

        Note that you need to implement this method directly on its child. Essentially,
        each type of Siamese has an unique step.

        Args:
            x (tf.Tensor): Tensor containing samples.
            y (tf.Tensor): Tensor containing labels.

        Raises:
            NotImplementedError.

        """

        raise NotImplementedError

    def fit(self, batches, epochs=100):
        """Method that trains the model over training batches.

        Note that you need to implement this method directly on its child. Essentially,
        each type of Siamese may use a distinct type of dataset.

        Args:
            batches (Dataset): Batches of tuples holding training samples and labels.
            epochs (int): Maximum number of epochs.

        Raises:
            NotImplementedError.

        """

        raise NotImplementedError

    def evaluate(self, batches):
        """Method that evaluates the model over validation or testing batches.

        Note that you need to implement this method directly on its child. Essentially,
        each type of Siamese may use a distinct type of dataset.

        Args:
            batches (Dataset): Batches of tuples holding validation / testing samples and labels.

        Raises:
            NotImplementedError.

        """

        raise NotImplementedError

    def predict(self, x):
        """Method that performs a forward pass over samples and returns the network's output.

        Note that you need to implement this method directly on its child. Essentially,
        each type of Siamese may predict in a different way.

        Args:
            x (tf.Tensor): Tensor containing samples.

        Raises:
            NotImplementedError.

        """

        raise NotImplementedError

    def extract_embeddings(self, x):
        """Method that extracts embeddings by performing a forward pass
        over the base architecture (embedder).

        Args:
            x (np.array, tf.Tensor): Array or tensor containing the inputs to be embedded.
            input_shape (tuple): Shape of the input layer.

        Returns:
            A tensor containing the embedded inputs.

        """

        # Converts the input to a tensor
        x = tf.convert_to_tensor(x)

        # Passes down through the embedder
        x = self.B(x)

        return x



"""Balanced- and random-pair datasets.
"""

import numpy as np
import tensorflow as tf

import dualing.utils.constants as c
import dualing.utils.exception as e
import dualing.utils.logging as l
from dualing.core import Dataset

logger = l.get_logger(__name__)


class BalancedPairDataset(Dataset):
    """A BalancedPairDataset class is responsible for implementing a dataset
    that creates balanced pairs of data, as well as their similarity (1) or dissimilarity (0).

    """

    def __init__(self, data, labels, n_pairs=2, batch_size=1, input_shape=None, normalize=(0, 1),
                 shuffle=True, seed=0):
        """Initialization method.

        Args:
            data (np.array): Array of samples.
            labels (np.array): Array of labels.
            n_pairs (int): Number of pairs.
            batch_size (int): Batch size.
            input_shape (tuple): Shape of the reshaped array.
            normalize (tuple): Normalization bounds.
            shuffle (bool): Whether data should be shuffled or not.
            seed (int): Provides deterministic traits when using `random` module.

        """

        logger.info('Overriding class: Dataset -> BalancedPairDataset.')

        # Overrides its parent class with any custom arguments if needed
        super(BalancedPairDataset, self).__init__(
            batch_size, input_shape, normalize, shuffle, seed)

        # Tries to assert the following statement
        try:
            # Checks if supplied labels are not equal
            assert not np.all(labels == labels[0])

        # If statement is not valid
        except:
            # Raises an error
            raise e.ValueError('`labels` should have distinct values')

        # Amount of pairs
        self.n_pairs = n_pairs

        # Pre-processes the data
        data = self.preprocess(data)

        # Creates pairs of data and labels
        pairs = self.create_pairs(data, labels)

        # Builds up the class
        self._build(pairs)

        logger.info('Class overrided.')

    @property
    def n_pairs(self):
        """int: Amount of pairs.

        """

        return self._n_pairs

    @n_pairs.setter
    def n_pairs(self, n_pairs):
        if not isinstance(n_pairs, int):
            raise e.TypeError('`n_pairs` should be a integer')

        self._n_pairs = n_pairs

    @property
    def batches(self):
        """tf.data.Dataset: Batches of data (samples, labels).

        """

        return self._batches

    @batches.setter
    def batches(self, batches):
        if not isinstance(batches, tf.data.Dataset):
            raise e.TypeError('`batches` should be a tf.data.Dataset')

        self._batches = batches

    def create_pairs(self, data, labels):
        """Creates balanced pairs from data and labels.

        Args:
            data (np.array): Array of samples.
            labels (np.array): Array of labels.

        Returns:
            Tuple containing pairs of samples along their labels.

        """

        logger.debug('Creating pairs ...')

        # Defines the number of samples
        n_samples = data.shape[0]

        # Divides equally the number of pairs
        n_pairs = self.n_pairs // 2

        # Defines the positive lists
        x1_p, x2_p, y_p = [], [], []

        # Defines the negative lists
        x1_n, x2_n, y_n = [], [], []

        # Iterates until both positive and negative pairs
        while len(y_p) < n_pairs or len(y_n) < n_pairs:
            # Samples two random indexes
            idx = tf.random.uniform([2], maxval=n_samples, dtype='int32')

            # If labels are equal on the particular indexes
            if tf.equal(tf.gather(labels, idx[0]), tf.gather(labels, idx[1])):
                # Appends positive data to `x1`
                x1_p.append(tf.gather(data, idx[0]))

                # Appends positive data to `x2`
                x2_p.append(tf.gather(data, idx[1]))

                # Appends positive label to `y`
                y_p.append(1.0)

            # If labels are not equal on the particular indexes
            else:
                # Appends negative data to `x1`
                x1_n.append(tf.gather(data, idx[0]))

                # Appends negative data to `x2`
                x2_n.append(tf.gather(data, idx[1]))

                # Appends negative label to `y`
                y_n.append(0.0)

        # Merges the positive and negative `x1`
        x1 = x1_p[:n_pairs] + x1_n[:n_pairs]

        # Merges the positive and negative `x2`
        x2 = x2_p[:n_pairs] + x2_n[:n_pairs]

        # Merges the positive and negative labels
        y = y_p[:n_pairs] + y_n[:n_pairs]

        logger.debug('Pairs: %s.', self.n_pairs)

        return x1, x2, y

    def _build(self, pairs):
        """Builds the class.

        Args:
            pairs (tuple): Pairs of samples along their labels.

        """

        # Checks if data should be shuffled
        if self.shuffle:
            # Creates dataset from shuffled and batched data
            self.batches = tf.data.Dataset.from_tensor_slices(
                pairs).shuffle(c.BUFFER_SIZE).batch(self.batch_size)

        # If data should not be shuffled
        else:
            # Creates dataset from batched data
            self.batches = tf.data.Dataset.from_tensor_slices(
                pairs).batch(self.batch_size)


class RandomPairDataset(Dataset):
    """A RandomPairDataset class is responsible for implementing a dataset that
    randomly creates pairs of data, as well as their similarity (1) or not (0).

    """

    def __init__(self, data, labels, batch_size=1, input_shape=None, normalize=(0, 1), seed=0):
        """Initialization method.

        Args:
            data (np.array): Array of samples.
            labels (np.array): Array of labels.
            batch_size (int): Batch size.
            input_shape (tuple): Shape of the reshaped array.
            normalize (tuple): Normalization bounds.
            seed (int): Provides deterministic traits when using `random` module.

        """

        logger.info('Overriding class: Dataset -> RandomPairDataset.')

        # Overrides its parent class with any custom arguments if needed
        super(RandomPairDataset, self).__init__(
            batch_size, input_shape, normalize, False, seed)

        # Pre-processes the data
        data = self.preprocess(data)

        # Creates pairs of data and labels
        pairs = self.create_pairs(data, labels)

        # Builds up the class
        self._build(pairs)

        logger.info('Class overrided.')

    @property
    def batches(self):
        """tf.data.Dataset: Batches of data (samples, labels).

        """

        return self._batches

    @batches.setter
    def batches(self, batches):
        if not isinstance(batches, tf.data.Dataset):
            raise e.TypeError('`batches` should be a tf.data.Dataset')

        self._batches = batches

    def _build(self, pairs):
        """Builds the class.

        Args:
            pairs (tuple): Pairs of samples along their labels.

        """

        # Creates batches from tensor slices
        self.batches = tf.data.Dataset.from_tensor_slices(
            pairs).batch(self.batch_size)

    def create_pairs(self, data, labels):
        """Creates random pairs from data and labels.

        Args:
            data (np.array): Array of samples.
            labels (np.array): Array of labels.

        Returns:
            Tuple containing pairs of samples along their labels.

        """

        logger.debug('Creating pairs ...')

        # Defines the number of samples
        n_samples = data.shape[0]

        # Defines the number of possible pairs
        n_pairs = n_samples // 2

        # Randomly samples indexes
        indexes = tf.random.shuffle(tf.range(n_samples))

        # Gathers samples
        x1, x2 = tf.gather(data, indexes[:n_pairs]), tf.gather(
            data, indexes[n_pairs:])

        # Gathers samples
        y1, y2 = tf.gather(labels, indexes[:n_pairs]), tf.gather(
            labels, indexes[n_pairs:])

        # If labels are equal, it means that samples are similar
        y = tf.cast(tf.equal(y1, y2), 'float32')

        # Calculates the number of positive and negative pairs
        n_pos_pairs = tf.math.count_nonzero(y)
        n_neg_pairs = y.shape[0] - n_pos_pairs

        logger.debug('Positive pairs: %s | Negative pairs: %s',
                     n_pos_pairs, n_neg_pairs)

        return (x1, x2, y)



"""Cross-Entropy Siamese Network.
"""

import tensorflow as tf

import dualing.utils.logging as l
from dualing.core import BinaryCrossEntropy, Siamese

logger = l.get_logger(__name__)


class CrossEntropySiamese(Siamese):
    """A CrossEntropySiamese class is responsible for implementing the
    cross-entropy version of Siamese Neural Networks.

    References:
        G. Koch, R. Zemel and R. Salakhutdinov.
        Siamese neural networks for one-shot image recognition.
        ICML Deep Learning Workshop (2015).

    """

    def __init__(self, base, name=''):
        """Initialization method.

        Args:
            base (Base): Twin architecture.
            name (str): Naming identifier.

        """

        logger.info('Overriding class: Siamese -> CrossEntropySiamese.')

        # Overrides its parent class with any custom arguments if needed
        super(CrossEntropySiamese, self).__init__(base, name=name)

        # Defines the output layer
        self.o = tf.keras.layers.Dense(1, activation='sigmoid')

        logger.info('Class overrided.')

    def compile(self, optimizer):
        """Method that builds the network by attaching optimizer, loss and metrics.

        Args:
            optimizer (tf.keras.optimizers): Optimization algorithm.

        """

        # Creates an optimizer object
        self.optimizer = optimizer

        # Defines the loss function
        self.loss = BinaryCrossEntropy()

        # Defines the accuracy function
        self.acc = tf.keras.metrics.binary_accuracy

        # Defines the loss metric
        self.loss_metric = tf.metrics.Mean(name='loss')

        # Defines the accuracy metric
        self.acc_metric = tf.metrics.Mean(name='acc')

    @tf.function
    def step(self, x1, x2, y):
        """Method that performs a single batch optimization step.

        Args:
            x1 (tf.Tensor): Tensor containing first samples from input pairs.
            x2 (tf.Tensor): Tensor containing second samples from input pairs.
            y (tf.Tensor): Tensor containing labels (1 for similar, 0 for dissimilar).

        """

        # Uses tensorflow's gradient
        with tf.GradientTape() as tape:
            # Passes the first sample through the network
            z1 = self.B(x1)

            # Passes the second sample through the network
            z2 = self.B(x2)

            # Passes the distance through sigmoid activation and removes last dimension
            y_pred = tf.squeeze(self.o(tf.abs(z1 - z2)), -1)

            # Calculates the loss
            loss = self.loss(y, y_pred)

            # Calculates the accuracy
            acc = self.acc(y, y_pred)

        # Calculates the gradients for each training variable based on the loss function
        gradients = tape.gradient(loss, self.B.trainable_variables)

        # Applies the gradients using an optimizer
        self.optimizer.apply_gradients(
            zip(gradients, self.B.trainable_variables))

        # Updates the metrics' states
        self.loss_metric.update_state(loss)
        self.acc_metric.update_state(acc)

    def fit(self, batches, epochs=100):
        """Method that trains the model over training batches.

        Args:
            batches (BalancedPairDataset, RandomPairDataset): Batches of tuples holding
                training samples and labels.
            epochs (int): Maximum number of epochs.

        """

        logger.info('Fitting model ...')

        # Gathers the amount of batches
        n_batches = tf.data.experimental.cardinality(batches).numpy()

        # Iterates through all epochs
        for epoch in range(epochs):
            logger.info('Epoch %d/%d', epoch+1, epochs)

            # Resets metrics' states
            self.loss_metric.reset_states()
            self.acc_metric.reset_states()

            # Defines a customized progress bar
            b = tf.keras.utils.Progbar(
                n_batches, stateful_metrics=['loss', 'acc'])

            # Iterates through all batches
            for (x1_batch, x2_batch, y_batch) in batches:
                # Performs the optimization step
                self.step(x1_batch, x2_batch, y_batch)

                # Adds corresponding values to the progress bar
                b.add(1, values=[('loss', self.loss_metric.result()),
                                 ('acc', self.acc_metric.result())])

            logger.file(
                f'Loss: {self.loss_metric.result()} | Accuracy: {self.acc_metric.result()}')

    def evaluate(self, batches):
        """Method that evaluates the model over validation or testing batches.

        Args:
            batches (BalancedPairDataset, RandomPairDataset): Batches of tuples holding
                validation / testing samples and labels.

        """

        logger.info('Evaluating model ...')

        # Gathers the amount of batches
        n_batches = tf.data.experimental.cardinality(batches).numpy()

        # Resets metrics' states
        self.loss_metric.reset_states()
        self.acc_metric.reset_states()

        # Defines a customized progress bar
        b = tf.keras.utils.Progbar(n_batches, stateful_metrics=['val_loss', 'val_acc'])

        # Iterates through all batches
        for (x1_batch, x2_batch, y_batch) in batches:
            # Performs the prediction
            y_pred = self.predict(x1_batch, x2_batch)

            # Calculates the loss
            loss = self.loss(y_batch, y_pred)

            # Calculates the accuracy
            acc = self.acc(y_batch, y_pred)

            # Updates the metrics' states
            self.loss_metric.update_state(loss)
            self.acc_metric.update_state(acc)

            # Adds corresponding values to the progress bar
            b.add(1, values=[('val_loss', self.loss_metric.result()),
                             ('val_acc', self.acc_metric.result())])

        logger.file(
            f'Val Loss: {self.loss_metric.result()} | Val Accuracy: {self.acc_metric.result()}')

    def predict(self, x1, x2):
        """Method that performs a forward pass over samples and returns the network's output.

        Args:
            x1 (tf.Tensor): Tensor containing first samples from input pairs.
            x2 (tf.Tensor): Tensor containing second samples from input pairs.

        Returns:
            The similarity score between samples `x1` and `x2`.

        """

        # Passes the first sample through the network
        z1 = self.B(x1)

        # Passes the second sample through the network
        z2 = self.B(x2)

        # Passes the distance through sigmoid activation and removes last dimension
        y_pred = tf.squeeze(self.o(tf.abs(z1 - z2)), -1)

        return y_pred



"""Losses.
"""

import tensorflow as tf
import tensorflow_addons as tfa


class BinaryCrossEntropy:
    """A BinaryCrossEntropy class defines the binary cross-entropy loss.

    """

    def __call__(self, y_true, y_pred):
        """Method that holds vital information whenever this class is called.

        Args:
            y_true (tf.Tensor): Tensor containing the true labels.
            y_pred (tf.Tensor): Tensor containing the predictions, e.g., similar or dissimilar.

        Returns:
            Binary cross-entropy loss.

        """

        return tf.keras.losses.binary_crossentropy(y_true, y_pred)


class ContrastiveLoss:
    """A ContrastiveEntropy class defines the contrastive loss.

    """

    def __call__(self, y_true, y_pred, margin=1.0):
        """Method that holds vital information whenever this class is called.

        Args:
            y_true (tf.Tensor): Tensor containing the true labels.
            y_pred (tf.Tensor): Tensor containing the predictions, e.g., distance.
            margin (float): Radius around the embedding space.

        Returns:
            Contrastive loss.

        """

        return tfa.losses.contrastive_loss(y_true, y_pred, margin)


class TripletHardLoss:
    """A TripletHardLoss class defines the triplet loss with hard negative mining.

    """

    def __call__(self, y_true, y_pred, margin=1.0, soft=False, distance_metric='L2'):
        """Method that holds vital information whenever this class is called.

        Args:
            y_true (tf.Tensor): Tensor containing the true labels.
            y_pred (tf.Tensor): Tensor containing the predictions, e.g., embeddings.
            margin (float): Radius around the embedding space.
            soft (bool): Whether network should use soft margin or not.
            distance_metric (str): Distance metric.

        Returns:
            Triplet loss with hard negative mining.

        """

        return tfa.losses.triplet_hard_loss(y_true, y_pred, margin, soft, distance_metric)


class TripletSemiHardLoss:
    """A TripletSemiHardLoss class defines the triplet loss with semi-hard negative mining.

    """

    def __call__(self, y_true, y_pred, margin=1.0, soft=None, distance_metric='L2'):
        """Method that holds vital information whenever this class is called.

        Args:
            y_true (tf.Tensor): Tensor containing the true labels.
            y_pred (tf.Tensor): Tensor containing the predictions, e.g., embeddings.
            margin (float): Radius around the embedding space.
            soft (None): Only for retro-compatibility.
            distance_metric (str): Distance metric.

        Returns:
            Triplet loss with semi-hard negative mining.

        """

        return tfa.losses.triplet_semihard_loss(y_true, y_pred, margin, distance_metric)



"""Contrastive Loss Siamese Network.
"""

import tensorflow as tf

import dualing.utils.exception as e
import dualing.utils.logging as l
from dualing.core import ContrastiveLoss, Siamese

logger = l.get_logger(__name__)


class ContrastiveSiamese(Siamese):
    """A ContrastiveSiamese class is responsible for implementing the
    contrastive version of Siamese Neural Networks.

    References:
        I. Melekhov, J. Kannala and E. Rahtu. Siamese network features for image matching.
        23rd International Conference on Pattern Recognition (2016).

    """

    def __init__(self, base, margin=1.0, distance_metric='L2', name=''):
        """Initialization method.

        Args:
            base (Base): Twin architecture.
            margin (float): Radius around the embedding space.
            distance_metric (str): Distance metric.
            name (str): Naming identifier.

        """

        logger.info('Overriding class: Siamese -> ContrastiveSiamese.')

        # Overrides its parent class with any custom arguments if needed
        super(ContrastiveSiamese, self).__init__(base, name=name)

        # Radius around embedding space
        self.margin = margin

        # Distance metric
        self.distance = distance_metric

        logger.info('Class overrided.')

    @property
    def margin(self):
        """float: Radius around the embedding space.

        """

        return self._margin

    @margin.setter
    def margin(self, margin):
        if not isinstance(margin, float):
            raise e.TypeError('`margin` should be a float')
        if margin <= 0:
            raise e.ValueError('`margin` should be greater than 0')

        self._margin = margin

    @property
    def distance(self):
        """str: Distance metric.

        """

        return self._distance

    @distance.setter
    def distance(self, distance):
        if distance not in ['L1', 'L2', 'angular']:
            raise e.ValueError('`distance` should be `L1`, `L2` or `angular`')

        self._distance = distance

    def compile(self, optimizer):
        """Method that builds the network by attaching optimizer, loss and metrics.

        Args:
            optimizer (tf.keras.optimizers): Optimization algorithm.

        """

        # Creates an optimizer object
        self.optimizer = optimizer

        # Defines the loss function
        self.loss = ContrastiveLoss()

        # Defines the loss metric
        self.loss_metric = tf.metrics.Mean(name='loss')

    @tf.function
    def step(self, x1, x2, y):
        """Method that performs a single batch optimization step.

        Args:
            x1 (tf.Tensor): Tensor containing first samples from input pairs.
            x2 (tf.Tensor): Tensor containing second samples from input pairs.
            y (tf.Tensor): Tensor containing labels (1 for similar, 0 for dissimilar).

        """

        # Uses tensorflow's gradient
        with tf.GradientTape() as tape:
            # Passes the first sample through the network
            z1 = self.B(x1)

            # Passes the second sample through the network
            z2 = self.B(x2)

            # If distance is supposed to be L1
            if self.distance == 'L1':
                # Calculates the L1 distance
                y_pred = tf.math.sqrt(tf.linalg.norm(z1 - z2, axis=1))

            # If distance is supposed to be L2
            elif self.distance == 'L2':
                # Calculates the L2 distance
                y_pred = tf.linalg.norm(z1 - z2, axis=1)

            # If distance is supposed to be angular
            elif self.distance == 'angular':
                # Calculates the angular distance
                y_pred = tf.keras.losses.cosine_similarity(z1, z2)

            # Calculates the loss
            loss = self.loss(y, y_pred, self.margin)

        # Calculates the gradients for each training variable based on the loss function
        gradients = tape.gradient(loss, self.B.trainable_variables)

        # Applies the gradients using an optimizer
        self.optimizer.apply_gradients(
            zip(gradients, self.B.trainable_variables))

        # Updates the metrics' states
        self.loss_metric.update_state(loss)

    def fit(self, batches, epochs=100):
        """Method that trains the model over training batches.

        Args:
            batches (BalancedPairDataset, RandomPairDataset): Batches of tuples holding
                training samples and labels.
            epochs (int): Maximum number of epochs.

        """

        logger.info('Fitting model ...')

        # Gathers the amount of batches
        n_batches = tf.data.experimental.cardinality(batches).numpy()

        # Iterates through all epochs
        for epoch in range(epochs):
            logger.info('Epoch %d/%d', epoch+1, epochs)

            # Resets metrics' states
            self.loss_metric.reset_states()

            # Defines a customized progress bar
            b = tf.keras.utils.Progbar(n_batches, stateful_metrics=['loss'])

            # Iterates through all batches
            for (x1_batch, x2_batch, y_batch) in batches:
                # Performs the optimization step
                self.step(x1_batch, x2_batch, y_batch)

                # Adds corresponding values to the progress bar
                b.add(1, values=[('loss', self.loss_metric.result())])

            logger.file(f'Loss: {self.loss_metric.result()}')

    def evaluate(self, batches):
        """Method that evaluates the model over validation or testing batches.

        Args:
            batches (BalancedPairDataset, RandomPairDataset): Batches of tuples holding
                validation / testing samples and labels.

        """

        logger.info('Evaluating model ...')

        # Gathers the amount of batches
        n_batches = tf.data.experimental.cardinality(batches).numpy()

        # Resets metrics' states
        self.loss_metric.reset_states()

        # Defines a customized progress bar
        b = tf.keras.utils.Progbar(n_batches, stateful_metrics=['val_loss'])

        # Iterates through all batches
        for (x1_batch, x2_batch, y_batch) in batches:
            # Performs the prediction
            y_pred = self.predict(x1_batch, x2_batch)

            # Calculates the loss
            loss = self.loss(y_batch, y_pred, self.margin)

            # Updates the metrics' states
            self.loss_metric.update_state(loss)

            # Adds corresponding values to the progress bar
            b.add(1, values=[('val_loss', self.loss_metric.result())])

        logger.file(f'Val Loss: {self.loss_metric.result()}')

    def predict(self, x1, x2):
        """Method that performs a forward pass over samples and returns the network's output.

        Args:
            x1 (tf.Tensor): Tensor containing first samples from input pairs.
            x2 (tf.Tensor): Tensor containing second samples from input pairs.

        Returns:
            The distance between samples `x1` and `x2`.

        """

        # Passes the first sample through the network
        z1 = self.B(x1)

        # Passes the second sample through the network
        z2 = self.B(x2)

        # If distance is supposed to be L1
        if self.distance == 'L1':
            # Calculates the L1 distance
            y_pred = tf.math.sqrt(tf.linalg.norm(z1 - z2, axis=1))

        # If distance is supposed to be L2
        elif self.distance == 'L2':
            # Calculates the L2 distance
            y_pred = tf.linalg.norm(z1 - z2, axis=1)

        # If distance is supposed to be angular
        elif self.distance == 'angular':
            # Calculates the angular distance
            y_pred = tf.keras.losses.cosine_similarity(z1, z2)

        return y_pred



"""Exceptions.
"""

import dualing.utils.logging as l

logger = l.get_logger(__name__)


class Error(Exception):
    """A generic Error class derived from Exception.

    Essentially, it gets the class and message and logs the error to the logger.

    """

    def __init__(self, cls, msg):
        """Initialization method.

        Args:
            cls (str): Class identifier.
            msg (str): Message to be logged.

        """

        # Override its parent class
        super(Error, self).__init__()

        # Logs the error in a formatted way
        logger.error('%s: %s.', cls, msg)


class ArgumentError(Error):
    """An ArgumentError class for logging errors related to wrong number of provided arguments.

    """

    def __init__(self, error):
        """Initialization method.

        Args:
            error (str): Error message to be logged.

        """

        # Override its parent class with class name and error message
        super(ArgumentError, self).__init__('ArgumentError', error)


class BuildError(Error):
    """A BuildError class for logging errors related to classes not being built.

    """

    def __init__(self, error):
        """Initialization method.

        Args:
            error (str): Error message to be logged.

        """

        # Override its parent class with class name and error message
        super(BuildError, self).__init__('BuildError', error)


class SizeError(Error):
    """A SizeError class for logging errors related to wrong length or size of variables.

    """

    def __init__(self, error):
        """Initialization method.

        Args:
            error (str): Error message to be logged.

        """

        # Override its parent class with class name and error message
        super(SizeError, self).__init__('SizeError', error)


class TypeError(Error):
    """A TypeError class for logging errors related to wrong type of variables.

    """

    def __init__(self, error):
        """Initialization method.

        Args:
            error (str): Error message to be logged.

        """

        # Override its parent class with class name and error message
        super(TypeError, self).__init__('TypeError', error)


class ValueError(Error):
    """A ValueError class for logging errors related to wrong value of variables.

    """

    def __init__(self, error):
        """Initialization method.

        Args:
            error (str): Error message to be logged.

        """

        # Override its parent class with class name and error message
        super(ValueError, self).__init__('ValueError', error)



from setuptools import find_packages, setup

with open('README.md', 'r') as f:
    long_description = f.read()

setup(name='dualing',
      version='1.0.1',
      description='Dual-based Neural Learning',
      long_description=long_description,
      long_description_content_type='text/markdown',
      author='Gustavo Rosa',
      author_email='gth.rosa@uol.com.br',
      url='https://github.com/gugarosa/dualing',
      license='MIT',
      install_requires=['coverage>=5.1',
                        'matplotlib>=3.2.2',
                        'pylint>=2.5.2',
                        'pytest>=5.4.2',
                        'tensorflow==2.2.0',
                        'tensorflow-addons>=0.10.0'
                        ],
      extras_require={
          'tests': ['coverage',
                    'pytest',
                    'pytest-pep8',
                    ],
      },
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Intended Audience :: Developers',
          'Intended Audience :: Education',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Topic :: Software Development :: Libraries',
          'Topic :: Software Development :: Libraries :: Python Modules'
      ],
      packages=find_packages())



"""Multi-Layer Perceptron.
"""

from tensorflow.keras.layers import Dense

import dualing.utils.logging as l
from dualing.core import Base

logger = l.get_logger(__name__)


class MLP(Base):
    """An MLP class stands for a Multi-Layer Perceptron implementation.

    """

    def __init__(self, n_hidden=[128]):
        """Initialization method.

        """

        logger.info('Overriding class: Base -> MLP.')

        # Overrides its parent class with any custom arguments if needed
        super(MLP, self).__init__(name='mlp')

        # Fully-connected layers
        self.fc = [Dense(units) for units in n_hidden]

        logger.info('Class overrided.')
        logger.debug('Layers: %d | Units: %d', len(n_hidden), n_hidden)

    def call(self, x):
        """Method that holds vital information whenever this class is called.

        Args:
            x (tf.Tensor): Tensor containing the input sample.

        Returns:
            The layer's outputs.

        """

        # Iterates through all fully-connected layers
        for fc in self.fc:
            # Passes down through the layer
            x = fc(x)

        return x
